
#include <cyg/infra/cyg_type.h>
#include <cyg/infra/testcase.h>         // Test macros
#include <cyg/infra/cyg_ass.h>          // Assertion macros
#include <cyg/infra/diag.h>             // Diagnostic output

#include <cyg/hal/hal_arch.h>           // CYGNUM_HAL_STACK_SIZE_TYPICAL
#include <cyg/kernel/kapi.h>

#include <cyg/io/spi.h>                 // Common SPI API
#include <cyg/io/spi_pl022.h>           // Pl022 data structures

//---------------------------------------------------------------------------
// Thread data structures.

cyg_uint8 stack [CYGNUM_HAL_STACK_SIZE_TYPICAL];
cyg_thread thread_data;
cyg_handle_t thread_handle;

//---------------------------------------------------------------------------
// SPI loopback device driver data structures.

cyg_spi_arm_pl022_device_t loopback_device = {
    .spi_device.spi_bus = &cyg_spi_arm_pl022_bus.spi_bus,
    .dev_num = -1,                      // No chip selects. 
    .cl_pol = 1,
    .cl_pha = 1,
    .cl_brate = 8000000,                // Nominal 8Mhz.
    .cs_up_udly = 1,
    .cs_dw_udly = 1,
    .tr_bt_udly = 1,
    .bus_loopback = true,
    .bus_16bit = false,
    .init = false                       // Default to uninitialised state
};

//---------------------------------------------------------------------------
// Run single loopback transaction using simple transfer API call.

void run_test_1 (void)
{
    char tx_data [] = "Testing, testing, 12, 123.";
    char rx_data [sizeof(tx_data)];

    diag_printf ("Test 1 : Simple transfer test.\n");
    cyg_spi_transfer (&loopback_device.spi_device, false, sizeof (tx_data), 
        (const cyg_uint8*) &tx_data[0], (cyg_uint8*) &rx_data[0]);

    diag_printf ("    Tx data : %s\n", tx_data);
    diag_printf ("    Rx data : %s\n", rx_data);
}

//---------------------------------------------------------------------------
// Run two loopback transactions using extended transfer API.

void run_test_2 (void)
{
    char tx_data1 [] = "Testing extended API...";
    char tx_data2 [] = "Testing extended API for a second transaction.";
    char rx_data1 [sizeof(tx_data1)];
    char rx_data2 [sizeof(tx_data2)];

    diag_printf ("Test 2 : Extended API test.\n");
    cyg_spi_transaction_begin (&loopback_device.spi_device);
    cyg_spi_transaction_transfer (&loopback_device.spi_device, false, sizeof (tx_data1), 
        (const cyg_uint8*) &tx_data1[0], (cyg_uint8*) &rx_data1[0], false);
    cyg_spi_transaction_transfer (&loopback_device.spi_device, false, sizeof (tx_data2), 
        (const cyg_uint8*) &tx_data2[0], (cyg_uint8*) &rx_data2[0], false);
    cyg_spi_transaction_end (&loopback_device.spi_device);

    diag_printf ("    Tx data 1 : %s\n", tx_data1);
    diag_printf ("    Rx data 1 : %s\n", rx_data1);
    diag_printf ("    Tx data 2 : %s\n", tx_data2);
    diag_printf ("    Rx data 2 : %s\n", rx_data2);
}

//---------------------------------------------------------------------------
// Run all PL022 SPI interface loopback tests.

void run_tests (void)
{
    diag_printf ("Running PL022 SPI driver loopback tests.\n");
    run_test_1(); 
    run_test_2(); 
    CYG_TEST_PASS_FINISH ("Loopback tests ran OK");
}

//---------------------------------------------------------------------------
// User startup - tests are run in their own thread.

void cyg_user_start(void)
{
    CYG_TEST_INIT();
    cyg_thread_create(
        10,                                   // Arbitrary priority
        (cyg_thread_entry_t*) run_tests,      // Thread entry point
        0,                                    // 
        "test_thread",                        // Thread name
        &stack[0],                            // Stack
        CYGNUM_HAL_STACK_SIZE_TYPICAL,        // Stack size
        &thread_handle,                       // Thread handle
        &thread_data                          // Thread data structure
    );
    cyg_thread_resume(thread_handle);
    cyg_scheduler_start();
}
