
#include <pkgconf/hal.h>
#include <pkgconf/io_spi.h>
#include <pkgconf/devs_spi_arm_pl022.h>

#include <cyg/hal/hal_io.h>
#include <cyg/hal/hal_if.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/hal/drv_api.h>

#include <cyg/infra/cyg_type.h>
#include <cyg/infra/cyg_ass.h>
#include <cyg/infra/diag.h>
#include <cyg/error/codes.h>

#include <cyg/io/spi.h>
#include <cyg/io/spi_pl022.h>
#include <cyg/io/spi_pl022_setup.h>

//-----------------------------------------------------------------------------
// Standard driver calls.

static void pl022_transaction_begin    (cyg_spi_device*);
static void pl022_transaction_transfer (cyg_spi_device*, cyg_bool, cyg_uint32, const cyg_uint8*, cyg_uint8*, cyg_bool);
static void pl022_transaction_tick     (cyg_spi_device*, cyg_bool, cyg_uint32);
static void pl022_transaction_end      (cyg_spi_device*);
static int  pl022_get_config           (cyg_spi_device*, cyg_uint32, void*, cyg_uint32*);
static int  pl022_set_config           (cyg_spi_device*, cyg_uint32, const void*, cyg_uint32*);

//-----------------------------------------------------------------------------
// Internally used functions, including ISR and DSR.

static cyg_uint32 pl022_ISR (cyg_vector_t vector, cyg_addrword_t data);
static void       pl022_DSR (cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);

static void do_transaction_intr (cyg_spi_device* device, cyg_bool tick_only,
    cyg_uint32 count, const cyg_uint8* tx_data, cyg_uint8* rx_data, cyg_bool drop_cs);

static void       write_to_fifo  (cyg_spi_arm_pl022_bus_t* pl022_bus, cyg_uint32 count);
static cyg_uint32 read_from_fifo (cyg_spi_arm_pl022_bus_t* pl022_bus);

//-----------------------------------------------------------------------------
// Specify the PL022 register locations and layout.

#define PL022_SSPCR0   (pl022_bus->reg_base + 0x00)
#define PL022_SSPCR1   (pl022_bus->reg_base + 0x04)
#define PL022_SSPDR    (pl022_bus->reg_base + 0x08)
#define PL022_SSPSR    (pl022_bus->reg_base + 0x0C)
#define PL022_SSPCPSR  (pl022_bus->reg_base + 0x10)
#define PL022_SSPIMSC  (pl022_bus->reg_base + 0x14)
#define PL022_SSPRIS   (pl022_bus->reg_base + 0x18)
#define PL022_SSPMIS   (pl022_bus->reg_base + 0x1C)
#define PL022_SSPICR   (pl022_bus->reg_base + 0x20)

#define PL022_SSPCR0_DSS_8BIT  0x0007
#define PL022_SSPCR0_DSS_16BIT 0x000F
#define PL022_SSPCR0_CPOL      0x0040
#define PL022_SSPCR0_CPHA      0x0080

#define PL022_SSPCR1_LBM       0x0001
#define PL022_SSPCR1_SSE       0x0002

#define PL022_SSPSR_TFE        0x0001
#define PL022_SSPSR_TNF        0x0002
#define PL022_SSPSR_RNE        0x0004
#define PL022_SSPSR_RFF        0x0008
#define PL022_SSPSR_BSY        0x0010

#define PL022_SSPINT_ROR       0x0001
#define PL022_SSPINT_RT        0x0002
#define PL022_SSPINT_RX        0x0004
#define PL022_SSPINT_TX        0x0008

//-----------------------------------------------------------------------------
// Configuration function for new PL022 bus.  This is called from the platform
// initialisation, which must also set up the associated GPIO pin mappings.

void cyg_spi_arm_pl022_bus_configure (cyg_spi_arm_pl022_bus_t* pl022_bus)
{
    // Fill in the function jump table.
    pl022_bus->spi_bus.spi_transaction_begin    = pl022_transaction_begin;
    pl022_bus->spi_bus.spi_transaction_transfer = pl022_transaction_transfer;
    pl022_bus->spi_bus.spi_transaction_tick     = pl022_transaction_tick;
    pl022_bus->spi_bus.spi_transaction_end      = pl022_transaction_end;
    pl022_bus->spi_bus.spi_get_config           = pl022_get_config;
    pl022_bus->spi_bus.spi_set_config           = pl022_set_config;

    // Disable PL022, clearing all chip selects and outstanding interrupts.
    HAL_WRITE_UINT32 (PL022_SSPCR1, 0);
    HAL_WRITE_UINT32 (PL022_SSPCPSR, CYGNUM_DEVS_SPI_ARM_PL022_CLK_PRESCALE);
    HAL_WRITE_UINT32 (PL022_SSPIMSC, 0);
    HAL_WRITE_UINT32 (PL022_SSPICR, PL022_SSPINT_ROR | PL022_SSPINT_RT);
    pl022_bus->spi_device_select(-1);
    pl022_bus->cs_up = false;

    // Select required interrupts.  Note that since we never have more than 8 frames in 
    // flight the RX FIFO will never overrun, so we can ignore PL022_SSPINT_ROR.
    HAL_WRITE_UINT32 (PL022_SSPIMSC, PL022_SSPINT_RT | PL022_SSPINT_RX); 

    // Hook in the ISR and DSR.
    cyg_drv_interrupt_create (CYGNUM_DEVS_SPI_ARM_PL022_INTERRUPT, 
        CYGNUM_DEVS_SPI_ARM_PL022_INTERRUPT_PRIORITY, (cyg_addrword_t) pl022_bus,   
        pl022_ISR, pl022_DSR, &pl022_bus->intr_handle, &pl022_bus->intr_data);
    cyg_drv_interrupt_attach (pl022_bus->intr_handle);

    // Initialise the transfer busy mutex and condition variable.
    cyg_drv_mutex_init (&pl022_bus->tfer_mutex);
    cyg_drv_cond_init (&pl022_bus->tfer_condvar, &pl022_bus->tfer_mutex);

    // Call upper layer bus init.
    CYG_SPI_BUS_COMMON_INIT(&pl022_bus->spi_bus);
}

//-----------------------------------------------------------------------------
// Start a new bus transaction.

static void pl022_transaction_begin (cyg_spi_device* device)
{
    cyg_spi_arm_pl022_bus_t* pl022_bus = (cyg_spi_arm_pl022_bus_t*) device->spi_bus;
    cyg_spi_arm_pl022_device_t* pl022_device = (cyg_spi_arm_pl022_device_t*) device;

    cyg_uint32 reg_data;

    // Place the clock divide data in the SCR field of SSPCR0.
    if (!pl022_device->init) {
        reg_data = CYGNUM_DEVS_SPI_ARM_PL022_CLK_BASE / 
            (pl022_device->cl_brate * CYGNUM_DEVS_SPI_ARM_PL022_CLK_PRESCALE);
        pl022_device->sspcr0_scr = (reg_data > 0xFF) ? 0xFF : (cyg_uint8) reg_data;
    }
    reg_data = ((cyg_uint32) pl022_device->sspcr0_scr) << 8;

    // Or in the other options for SSPCR0.
    if (pl022_device->bus_16bit)
        reg_data |= PL022_SSPCR0_DSS_16BIT;
    else
        reg_data |= PL022_SSPCR0_DSS_8BIT;
    if (pl022_device->cl_pol)
        reg_data |= PL022_SSPCR0_CPOL;
    if (pl022_device->cl_pha)
        reg_data |= PL022_SSPCR0_CPHA;

    HAL_WRITE_UINT32 (PL022_SSPCR0, reg_data);

    // Enable bus, with loopbacks if required.
    reg_data = PL022_SSPCR1_SSE;
    if (pl022_device->bus_loopback)
        reg_data |= PL022_SSPCR1_LBM;

    HAL_WRITE_UINT32 (PL022_SSPCR1, reg_data);

    // Bus width needs to be copied to the bus data structure so ISRs can access it.
    pl022_bus->bus_16bit = pl022_device->bus_16bit;
}

//-----------------------------------------------------------------------------
// Carry out a bus transfer.

static void pl022_transaction_transfer (cyg_spi_device* device, cyg_bool polled, 
    cyg_uint32 count, const cyg_uint8* tx_data, cyg_uint8* rx_data, cyg_bool drop_cs)
{
    cyg_spi_arm_pl022_bus_t* pl022_bus = (cyg_spi_arm_pl022_bus_t*) device->spi_bus;

    // Check for unsupported transactions.
    CYG_ASSERT (!polled,   "PL022 driver : Polled operation is not currently supported.");
    CYG_ASSERT (count > 0, "PL022 driver : Null transfer requested.");

    // We check that the buffers are half-word aligned and that count is a 
    // multiple of two in order to carry out the 16-bit transfer.
    if (pl022_bus->bus_16bit) {
        CYG_ASSERT (!(count & 1) && !((cyg_uint32) tx_data & 1) && !((cyg_uint32) rx_data & 1),   
            "PL022 driver : Misaligned data in 16-bit transfer.");
    }

    // Perform interrupt driven transfer.
    do_transaction_intr (device, false, count, tx_data, rx_data, drop_cs);
}

//-----------------------------------------------------------------------------
// Carry out a bus tick operation - this just pushes the required number of
// zeros onto the bus, leaving the chip select in its current state.

static void pl022_transaction_tick (cyg_spi_device* device, cyg_bool polled, cyg_uint32 count)
{
    cyg_spi_arm_pl022_bus_t* pl022_bus = (cyg_spi_arm_pl022_bus_t*) device->spi_bus;

    // Check for unsupported transactions.
    CYG_ASSERT (!polled,   "PL022 driver : Polled operation is not currently supported.");
    CYG_ASSERT (count > 0, "PL022 driver : Null transfer requested.");

    // We check that count is a multiple of two in order to carry out the 
    // 16-bit transfer.
    if (pl022_bus->bus_16bit) {
        CYG_ASSERT (!(count & 1),   
            "PL022 driver : Misaligned data in 16-bit transfer.");
    }

    // Perform interrupt driven transfer.
    do_transaction_intr (device, true, count, NULL, NULL, false);
}

//-----------------------------------------------------------------------------
// Complete the transaction.

static void pl022_transaction_end (cyg_spi_device* device)
{
    cyg_spi_arm_pl022_bus_t* pl022_bus = (cyg_spi_arm_pl022_bus_t*) device->spi_bus;
    cyg_spi_arm_pl022_device_t* pl022_device = (cyg_spi_arm_pl022_device_t*) device;

    // Ensure that the chip select is deasserted.
    if (pl022_bus->cs_up) {
        CYGACC_CALL_IF_DELAY_US (pl022_device->cs_dw_udly);
        pl022_bus->spi_device_select (-1);   
        pl022_bus->cs_up = false;
    }     

    // Disable the PL022.
    HAL_WRITE_UINT32 (PL022_SSPCR1, 0);
}

//-----------------------------------------------------------------------------
// Write the specified number of frames to the transmit FIFO.

static void write_to_fifo (cyg_spi_arm_pl022_bus_t* pl022_bus, cyg_uint32 count)
{
    cyg_uint32 i;

    // Fill the transmit FIFO with null data.
    if (pl022_bus->tx_data_ptr == NULL) {
        for (i = count; (i != 0) && (pl022_bus->tx_count != 0); i--) {
            HAL_WRITE_UINT32 (PL022_SSPDR, 0);
            pl022_bus->tx_count -= (pl022_bus->bus_16bit) ? 2 : 1; 
        }
    }

    // Fill the transmit FIFO with 16-bit data.
    else if (pl022_bus->bus_16bit) {
        for (i = count; (i != 0) && (pl022_bus->tx_count != 0); i--) {
            HAL_WRITE_UINT32 (PL022_SSPDR, (cyg_uint32) *((cyg_uint16*) pl022_bus->tx_data_ptr));
            pl022_bus->tx_data_ptr += 2;
            pl022_bus->tx_count -= 2; 
        }
    }

    // Fill the transmit FIFO with 8-bit data.
    else {
        for (i = count; (i != 0) && (pl022_bus->tx_count != 0); i--) {
            HAL_WRITE_UINT32 (PL022_SSPDR, (cyg_uint32) *(pl022_bus->tx_data_ptr++));
            pl022_bus->tx_count -= 1; 
        }
    }
}

//-----------------------------------------------------------------------------
// Drain the receive FIFO to the receive buffer, returning the number of frames
// received.

static cyg_uint32 read_from_fifo (cyg_spi_arm_pl022_bus_t* pl022_bus)
{
    cyg_uint32 pop_count = 0;
    cyg_uint32 rx_data, status;

    // Drain the receive FIFO, storing the data in the receive buffer if required.
    HAL_READ_UINT32 (PL022_SSPSR, status);
    while (status & PL022_SSPSR_RNE) {
        HAL_READ_UINT32 (PL022_SSPDR, rx_data);
        HAL_READ_UINT32 (PL022_SSPSR, status);
        pop_count++;

        // Skip writes for a null buffer.
        if (pl022_bus->rx_data_ptr == NULL) {
            pl022_bus->rx_count -= (pl022_bus->bus_16bit) ? 2 : 1; 
        }        

        // Deal with 16-bit data.
        else if (pl022_bus->bus_16bit) {
            *((cyg_uint16*) pl022_bus->rx_data_ptr) = (cyg_uint16) rx_data;
            pl022_bus->rx_data_ptr += 2;
            pl022_bus->rx_count -= 2; 
        }             

        // Deal with 8-bit data.
        else {
            *(pl022_bus->rx_data_ptr++) = (cyg_uint8) rx_data;
            pl022_bus->rx_count -= 1; 
        }             
    }
    return pop_count;
}

//-----------------------------------------------------------------------------
// Perform interrupt driven bus transfer or bus tick.

static void do_transaction_intr (cyg_spi_device* device, cyg_bool tick_only,
    cyg_uint32 count, const cyg_uint8* tx_data, cyg_uint8* rx_data, cyg_bool drop_cs)
{
    cyg_spi_arm_pl022_bus_t* pl022_bus = (cyg_spi_arm_pl022_bus_t*) device->spi_bus;
    cyg_spi_arm_pl022_device_t* pl022_device = (cyg_spi_arm_pl022_device_t*) device;

    // Ensure the chip select is asserted, inserting inter-transaction guard 
    // time if required.  Note that when ticking the device we do not touch the CS.
    if (!pl022_bus->cs_up && !tick_only) {
        CYGACC_CALL_IF_DELAY_US (pl022_device->tr_bt_udly);
        pl022_bus->spi_device_select (pl022_device->dev_num);
        pl022_bus->cs_up = true;
        CYGACC_CALL_IF_DELAY_US (pl022_device->cs_up_udly);
    }

    // Initialise the bus state
    pl022_bus->tx_count = count;
    pl022_bus->rx_count = count;
    pl022_bus->tx_data_ptr = tx_data;
    pl022_bus->rx_data_ptr = rx_data;

    // Fill the transmit buffer.
    write_to_fifo (pl022_bus, 8);

    // Sit back and wait for the ISR/DSR to signal completion.
    cyg_drv_mutex_lock (&pl022_bus->tfer_mutex);
    cyg_drv_dsr_lock ();
    cyg_drv_interrupt_unmask (CYGNUM_DEVS_SPI_ARM_PL022_INTERRUPT);

    while (pl022_bus->rx_count != 0)
        cyg_drv_cond_wait (&pl022_bus->tfer_condvar);

    cyg_drv_dsr_unlock ();
    cyg_drv_mutex_unlock (&pl022_bus->tfer_mutex);
        
    // Interrupts will have been disabled by the ISR/DSR, so all we need to do here
    // is deassert the chip select if required.
    if (drop_cs && !tick_only) {
        CYGACC_CALL_IF_DELAY_US (pl022_device->cs_dw_udly);
        pl022_bus->spi_device_select (-1);   
        pl022_bus->cs_up = false;
    }     
}

//-----------------------------------------------------------------------------
// Low level interrupt service routine.  This just clears down the interrupt
// condition by masking out the offending interrupts then schedules the DSR
// to do the heavy lifting.

static cyg_uint32 pl022_ISR (cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_mask (vector);
    cyg_drv_interrupt_acknowledge (vector);
    return (CYG_ISR_CALL_DSR | CYG_ISR_HANDLED);
}

//-----------------------------------------------------------------------------
// Delayed interrupt service routine.  Performs FIFO push and pop operations.

static void pl022_DSR (cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    cyg_spi_arm_pl022_bus_t* pl022_bus = (cyg_spi_arm_pl022_bus_t*) data;

    cyg_uint32 pop_count;

    // Drain the receive FIFO, storing the data in the receive buffer if required.
    pop_count = read_from_fifo (pl022_bus);

    // Refill the FIFO with the specified number of frames.
    write_to_fifo (pl022_bus, pop_count);

    // If more receive data is expected re-enable the interrupts and continue,
    // otherwise signal completion via the condition variable.
    if (pl022_bus->rx_count != 0) {
        cyg_drv_interrupt_unmask (vector);
    }
    else {
        cyg_drv_cond_signal (&pl022_bus->tfer_condvar);
    }
}

//-----------------------------------------------------------------------------
// Note that no dynamic configuration options are currently defined.

static int pl022_get_config (cyg_spi_device* dev, cyg_uint32 key, void* buf, cyg_uint32* len)
{
    return -EINVAL;
}

static int pl022_set_config (cyg_spi_device* dev, cyg_uint32 key, const void* buf, cyg_uint32* len)
{
    return -EINVAL;
}
